@@ -55,6 +55,11 @@ Usage:
 	If a variable name with a leading underscore ("_"), although it legal in
 	python, is passed to R, an RError will be raised.
 	
+Conversions:
+	Python -> R
+		None -> NULL, NaN -> NaN, Inf -> Inf
+	R -> Python (numpy)
+		NULL -> None, NA -> None, NaN -> None (NaN), Inf -> None (Inf)
 
 DEBUG model:
 	Since the child process (R) can be easily killed by any ocassional error in
@@ -79,9 +84,9 @@ import re
 import tempfile
 from types import *
 
-__version__ = '1.1.0'
+__version__ = '1.1.1'
 
-if sys.version < '2.3': # actually python 2.3 is required by tempfile.mkstemp !!!
+if sys.version < '2.3': # actually python >= 2.3 is required by tempfile.mkstemp used in this module !!!
 	set = frozenset = tuple
 	basestring = str
 elif sys.version < '2.4':
@@ -89,12 +94,18 @@ elif sys.version < '2.4':
 
 if sys.version < '3.0':
 	_mystr = _mybytes = lambda s:s
+	_in_py3 = False
 else:
 	from functools import reduce
 	long, basestring, unicode = int, str, str
 	_mybytes = lambda s:bytes(s, 'utf8') #'ascii')
 	_mystr = lambda s:str(s, 'utf8')
-
+	_in_py3 = True
+try:
+	import pandas
+	has_pandas = True
+except:
+	has_pandas = False
 try:
 	import numpy
 	has_numpy = True
@@ -158,6 +169,16 @@ def BoolStr(obj): return obj and 'TRUE' or 'FALSE'
 
 def ReprStr(obj): return repr(obj)
 
+if has_numpy:
+	def FloatStr(f):
+		if f is numpy.NaN or f is numpy.nan: return 'NaN' # or 'NA'
+		if has_pandas and pandas.isnull(f): return 'NaN'
+		if numpy.isposinf(f): return 'Inf'
+		if numpy.isneginf(f): return '-Inf'
+		return repr(f)
+else:
+	FloatStr = repr
+	
 def LongStr(obj):  
 	rtn = repr(obj)
 	if rtn[-1] == 'L': rtn = rtn[:-1]
@@ -166,6 +187,13 @@ def LongStr(obj):
 def ComplexStr(obj):
 	return repr(obj).replace('j', 'i')
 
+def UniStr(obj):
+	return repr(obj.encode('utf8'))
+
+def ByteStr(obj):
+	return repr(obj)[1:]
+	#return obj.decode()
+
 def SeqStr(obj, head='c(', tail=')'):
 	if not obj: return head + tail
 	# detect types
@@ -193,41 +221,53 @@ def SeqStr(obj, head='c(', tail=')'):
 def DictStr(obj):
 	return 'list(' + ','.join(['%s=%s' % (Str4R(a[0]), Str4R(a[1])) for a in obj.items()]) + ')'
 
+_tpdic = {'i':'as.integer(c(%s))', 'u':'as.integer(c(%s))', 'f':'as.double(c(%s))', 'c':'as.complex(c(%s))', 'b':'c(%s)', 'S':'c(%s)', 'a':'c(%s)', 'U':'c(%s)', 'V':'list(%s)'} # in order: (signed) integer, unsigned integer, float, complex, boolean, string, string, unicode, anything
+def getVec(ary): 
+	# used for objects from numpy and pandas
+	tp = ary.dtype.kind
+	if len(ary.shape) > 1:
+		ary = ary.reshape(reduce(lambda a,b=1:a*b, ary.shape))
+	ary = tp == 'b' and [a and 'TRUE' or 'FALSE' for a in ary] or ary.tolist()
+	if tp != 'V':
+		return _tpdic.get(tp, 'c(%s)') % repr(ary)[1:-1]
+	# record array
+	ary = list(map(SeqStr, ary)) # each record will be mapped to vector or list
+	return _tpdic.get(tp, 'list(%s)') % (', '.join(ary)) # use str here instead of repr since it has already been converted to str by SeqStr
+
+def NumpyNdarrayStr(obj):
+	shp = obj.shape
+	if len(shp) == 1: # to vector
+		tp = obj.dtype
+		if tp.kind != 'V': 
+			return getVec(obj)
+		# One-dimension record array will be converted to data.frame
+		def mapField(f):
+			ary = obj[f]
+			tp = ary.dtype.kind
+			return '"%s"=%s' % (f, _tpdic.get(tp, 'list(%s)') % repr(ary.tolist())[1:-1])
+		return 'data.frame(%s)' % (', '.join(map(mapField, tp.names)))
+	elif len(shp) == 2: # two-dimenstion array will be converted to matrix
+		return 'matrix(%s, nrow=%d, byrow=TRUE)' % (getVec(obj), shp[0])
+	else: # to array
+		dim = list(shp[-2:]) # row, col
+		dim.extend(shp[-3::-1])
+		newaxis = list(range(len(shp)))
+		newaxis[-2:] = [len(shp)-1, len(shp)-2]
+		return 'array(%s, dim=c(%s))' % (getVec(obj.transpose(newaxis)), repr(dim)[1:-1])
+
+def PandasSerieStr(obj):
+	return 'data.frame(%s=%s, row.names=%s)' % (obj.name, getVec(obj.values), getVec(obj.index))
+
+def PandasDataFrameStr(obj):
+	# DataFrame will be converted to data.frame, have to explicitly name columns
+	return 'data.frame(%s, row.names=%s)' % (', '.join(map(lambda a,b=obj:a+'='+getVec(obj[a]), obj)), getVec(obj.index))
+	s = ''
+	for col in obj:
+		s = s + col+'='+getVec(obj[col])+', '
+	#print 'data.frame(%s row.names=%s)' % (s, getVec(obj.index))
+	return 'data.frame(%s row.names=%s)' % (s, getVec(obj.index))
+
 def OtherStr(obj):
-	if has_numpy:
-		if isinstance(obj, numpy.ndarray):
-			shp = obj.shape
-			tpdic = {'i':'as.integer(c(%s))', 'u':'as.integer(c(%s))', 'f':'as.double(c(%s))', 'c':'as.complex(c(%s))', 'b':'c(%s)', 'S':'c(%s)', 'a':'c(%s)', 'U':'c(%s)', 'V':'list(%s)'} # in order: (signed) integer, unsigned integer, float, complex, boolean, string, string, unicode, anything
-
-			def getVec(ary):
-				tp = ary.dtype.kind
-				rlt = ary.reshape(reduce(lambda a,b=1:a*b, ary.shape))
-				rlt = tp == 'b' and [a and 'TRUE' or 'FALSE' for a in rlt] or rlt.tolist()
-				if tp != 'V':
-					return tpdic.get(tp, 'c(%s)') % repr(rlt)[1:-1]
-				# record array
-				rlt = list(map(SeqStr, rlt)) # each record will be mapped to vector or list
-				return tpdic.get(tp, 'list(%s)') % (', '.join(rlt)) # use str here instead of repr since it has already been converted to str by SeqStr
-
-			if len(shp) == 1: # to vector
-				tp = obj.dtype
-				if tp.kind != 'V': 
-					return getVec(obj)
-				# One-dimension record array will be converted to data.frame
-				def mapField(f):
-					ary = obj[f]
-					tp = ary.dtype.kind
-					return '"%s"=%s' % (f, tpdic.get(tp, 'list(%s)') % repr(ary.tolist())[1:-1])
-				return 'data.frame(%s)' % (', '.join(map(mapField, tp.names)))
-			elif len(shp) == 2: # two-dimenstion array will be converted to matrix
-				return 'matrix(%s, nrow=%d, byrow=TRUE)' % (getVec(obj), shp[0])
-			else: # to array
-				dim = list(shp[-2:]) # row, col
-				dim.extend(shp[-3::-1])
-				newaxis = list(range(len(shp)))
-				newaxis[-2:] = [len(shp)-1, len(shp)-2]
-				return 'array(%s, dim=c(%s))' % (getVec(obj.transpose(newaxis)), repr(dim)[1:-1])
-			# record array and char array
 	if hasattr(obj, '__iter__'): # for iterators
 		if hasattr(obj, '__len__') and len(obj) <= 10000:
 			return SeqStr(list(obj))
@@ -235,20 +275,34 @@ def OtherStr(obj):
 			return SeqStr(list(obj))
 	return repr(obj)
 
-base_tps = [type(None), bool, int, long, float, complex, str, unicode, list, tuple, set, frozenset, dict] # use type(None) instead of NoneType since the latter cannot be found in the types module in Python 3
+str_func = {NoneType:NoneStr, bool:BoolStr, long:LongStr, int:repr, float:FloatStr, complex:ComplexStr, unicode:UniStr, str:repr, list:SeqStr, tuple:SeqStr, set:SeqStr, frozenset:SeqStr, dict:DictStr} # str will override uncode in Python 3
+
+base_tps = [NoneType, bool, int, long, float, complex, str, unicode, list, tuple, set, frozenset, dict] # use type(None) instead of NoneType since the latter cannot be found in the types module in Python 3
+if has_numpy: 
+	str_func[numpy.ndarray] = NumpyNdarrayStr
+	base_tps.append(numpy.ndarray)
+if has_pandas:
+	str_func.update({pandas.Series:PandasSerieStr, pandas.DataFrame:PandasDataFrameStr})
+	base_tps.extend([pandas.Series, pandas.DataFrame])
 base_tps.reverse()
-str_func = {type(None):NoneStr, bool:BoolStr, long:LongStr, int:repr, float:repr, complex:ComplexStr, str:repr, unicode:repr, list:SeqStr, tuple:SeqStr, set:SeqStr, frozenset:SeqStr, dict:DictStr}
+
+if _in_py3:
+	base_tps.append(bytes)
+	str_func[bytes] = ByteStr
 
 def Str4R(obj):  
 	'''
 	convert a Python basic object into an R object in the form of string.
 	'''
 	#return str_func.get(type(obj), OtherStr)(obj) 
+	# for objects known by PypeR
 	if type(obj) in str_func:
 		return str_func[type(obj)](obj)
+	# for objects derived from basic data types
 	for tp in base_tps:
 		if isinstance(obj, tp):
 			return str_func[tp](obj)
+	# for any other objects
 	return OtherStr(obj)
 
 class RError(Exception):
@@ -257,33 +311,64 @@ class RError(Exception):
 	def __str__(self):
 		return repr(self.value)
 
-class R: # (object):
+class R(object): # "del r.XXX" fails on FePy-r7 (IronPython 1.1 on .NET 2.0.50727.42) if using old-style class
 	'''
 	A Python class to enclose an R process.
 	'''
-	__Rfun = r'''.getRvalue4Python__ <- function(x, use_dict=NULL) {
-	has_numpy <- %s
+	__Rfun = r'''.getRvalue4Python__ <- function(x, use_dict=NULL, has_numpy=FALSE, has_pandas=FALSE) {
 	if (has_numpy) {
 		headstr <- 'numpy.array('
-		tailstr <- ')'}
+		tailstr <- ')' }
 	else headstr <- tailstr <- ''
+	SpecialLocs <- function(x) {
+		rlt <- list()
+		if (!has_numpy) {
+			idx <- which(is.null(x) | is.na(x) | is.nan(x) | is.infinite(x))
+			if (length(idx) > 0) list$None <- idx
+			}
+		else {
+			idx <- which(is.null(x) | is.na(x))
+			if (length(idx) > 0) rlt$None <- idx
+			idx <- which(is.nan(x)) 
+			if (length(idx) > 0) rlt$numpy.NaN <- idx
+			idx <- which(is.infinite(x))
+			if (length(idx) > 0) {
+				v <- x[idx]
+				iidx <- which(v > 0)
+				if (length(iidx) > 0) rlt$numpy.Inf <- idx[iidx]
+				iidx <- which(v < 0)
+				if (length(iidx) > 0) rlt['-numpy.Inf'] <- idx[iidx]
+				}
+			}
+		return(rlt)
+		}
+	SpecialVals <- function(x, valoc) {
+		for (val in names(valoc)) x[valoc[[val]]] <- val
+		return(x)
+		}
 	NullStr <- function(x) 'None'
 	VectorStr <- function(x) {
 		#nms <- names(x)
 		#if (!is.null(nms) &&  length(nms)>0) return(ListStr(as.list(x)))
 		complx <- is.complex(x)
-		if (is.character(x)) x <- paste('"', x, '"', sep='')
+		special_locs <- SpecialLocs(x)
+		if (is.character(x)) {
+			x <- gsub('"', '\\\\"', x)
+			x <- paste('"', x, '"', sep='') }
 		else if (is.logical(x)) x <- ifelse(x, 'True', 'False')
-		if (length(x)==1) x <- paste(x) # convert to character, or use "gettext", "as.character"
-		else x <- paste(headstr, '[', paste(x, collapse=', '), ']', tailstr, sep='')
+		if (length(special_locs) > 0) x <- SpecialVals(x, special_locs)
+		if (length(x)==1) x <- paste(x) # convert to character using paste, "gettext", or "as.character"
+		else x <- paste(headstr, '[', paste(x, collapse=','), ']', tailstr, sep='')
 		if (complx) x <- gsub('i', 'j', x)
 		return(x) }
 	MatrixStr <- function(x) {
 		complx <- is.complex(x)
+		special_locs <- SpecialLocs(x)
 		if (is.character(x)) x <- matrix(paste('"', x, '"', sep=''), nrow=nrow(x))
 		else if (is.logical(x)) x <- ifelse(x, 'True', 'False')
-		x <- apply(x, 1, function(r) paste('[', paste(r, collapse=', '), ']', sep=''))
-		x <- paste(headstr, '[', paste(x, collapse=', '), ']', tailstr, sep='')
+		if (length(special_locs) > 0) x <- SpecialVals(x, special_locs)
+		x <- apply(x, 1, function(r) paste('[', paste(r, collapse=','), ']', sep=''))
+		x <- paste(headstr, '[', paste(x, collapse=','), ']', tailstr, sep='')
 		if (complx) x <- gsub('i', 'j', x)
 		return(x) }
 	ArrayStr <- function(x) {
@@ -294,18 +379,26 @@ class R: # (object):
 		# ndim >= 3
 		if (is.character(x)) x <- array(paste('"', x, '"', sep=''), dim=dim(x))
 		else if (is.logical(x)) x <- ifelse(x, 'True', 'False')
-		for (i in seq(ndim-1)) 
-			x <- apply(x, seq(dim(x))[-1], function(r) paste('[', paste(r, collapse=', '), ']', sep=''))
-		x <- paste(headstr, '[', paste(x, collapse=', '), ']', tailstr, sep='')
+		# do col first
+		x <- apply(x, seq(dim(x))[-2], function(r) paste('[', paste(r, collapse=','), ']', sep=''))
+		for (i in seq(ndim-2)) 
+			x <- apply(x, seq(dim(x))[-1], function(r) paste('[', paste(r, collapse=','), ']', sep=''))
+		x <- paste(headstr, '[', paste(x, collapse=','), ']', tailstr, sep='')
 		if (complx) x <- gsub('i', 'j', x)
 		return(x) }
 	DataFrameStr <- function(x) {
+		if (has_pandas) {
+			dimx <- dim(x)   # data frame, so length always > 1
+			if (dimx[2] == 1) return(PandasSeriesStr(x))
+			# if (dimx[2] > 1) return(PandasDFStr(x))
+			# if (length(dimx) == 3) return(PandasPanelStr(x)) not implemented yet
+			}
 		cnms <- colnames(x) # get column names
 		ctp <- list()
 		for (i in seq(x)) {
 			xi <- as.vector(x[[i]])
 			if (is.character(xi)) {
-				ctp[i] <- sprintf('("%%s", "|S%%d")', cnms[i], max(nchar(xi)) )
+				ctp[i] <- sprintf('("%s", "|S%d")', cnms[i], max(nchar(xi)) )
 				xi <- paste('"', xi, '"', sep='') }
 			else if (is.logical(xi)) {
 				xi <- ifelse(xi, 'True', 'False')
@@ -319,29 +412,58 @@ class R: # (object):
 			else if (is.complex(xi)) {
 				xi <- gsub('i', 'j', paste(xi))
 				ctp[i] <- paste('("', cnms[i], '", "<G")') }
+			special_locs <- SpecialLocs(xi)
+			if (length(special_locs) > 0) xi <- SpecialVals(xi, special_locs)
 			x[[i]] <- xi }
+		if (has_pandas) return(PandasDFStr(x))
 		x <- as.matrix(x)
-		x <- apply(x, 1, function(r) paste('(', paste(r, collapse=', '), ')', sep=''))
-		if (has_numpy) {
-			tailstr <- paste(', dtype=[', paste(ctp, collapse=', '), ']', tailstr, sep='')
-			}
-		x <- paste(headstr, '[', paste(x, collapse=', '), ']', tailstr, sep='')
+		x <- apply(x, 1, function(r) paste('(', paste(r, collapse=','), if(length(r)<2) ',)' else ')', sep=''))
+		if (has_numpy) tailstr <- paste(', dtype=[', paste(ctp, collapse=','), ']', tailstr, sep='')
+		x <- paste(headstr, '[', paste(x, collapse=','), ']', tailstr, sep='')
+		return(x) }
+	PandasSeriesStr <- function(x) {
+		headstr <- 'pandas.Series('
+		tailstr <- ')'
+		cnms <- colnames(x) # get column names
+		rnms <- row.names(x) # get row names
+		x <- VectorStr(x[,1])
+		tailstr <- paste(', name="', cnms[1], '"', tailstr, sep='')
+		if (! is.null(rnms)) {
+			tailstr <- paste(', index=["', paste(rnms, collapse='","'), '"]', tailstr, sep='') }
+		x <- paste(headstr, x, tailstr, sep='')
+		return(x) 
+		}
+	PandasDFStr <- function(x) {
+		headstr <- 'pandas.DataFrame('
+		tailstr <- ')'
+		cnms <- colnames(x) # get column names
+		rnms <- row.names(x) # get row names
+		#special_locs <- SpecialLocs(x)
+		x <- as.matrix(x)
+		#if (length(special_locs) > 0) x <- SpecialVals(x, special_locs)
+		x <- apply(x, 1, function(r) paste('(', paste(r, collapse=','), if(length(r)<2) ',)' else ')', sep=''))
+		tailstr <- paste(', columns=["', paste(cnms, collapse='","'), '"]', tailstr, sep='')
+		if (! is.null(rnms)) {
+			tailstr <- paste(', index=["', paste(rnms, collapse='","'), '"]', tailstr, sep='') }
+		x <- paste(headstr, '[', paste(x, collapse=','), ']', tailstr, sep='')
 		return(x) }
 	ListStr <- function(x) {
 		nms <- names(x) # get column names
 		x <- sapply(x, Str4Py)
+		return(zipVecWithName(x, nms))}
+	zipVecWithName <- function(x, nms) {
 		if (!is.null(nms) &&  length(nms)>0) {
 			nms <- paste('"', nms, '"', sep='')
 			x <- sapply(seq(nms), function(i) paste('(', nms[i], ',', x[i], ')') ) 
-			if (identical(use_dict, TRUE)) x <- paste('dict([', paste(x, collapse=', '), '])', sep='')
-			else if (identical(use_dict, FALSE))  x <- paste('[', paste(x, collapse=', '), ']', sep='')
+			if (identical(use_dict, TRUE)) x <- paste('dict([', paste(x, collapse=','), '])', sep='')
+			else if (identical(use_dict, FALSE))  x <- paste('[', paste(x, collapse=','), ']', sep='')
 			else { # should be NULL or something else
-				if (length(nms) != length(unique(nms))) x <- paste('[', paste(x, collapse=', '), ']', sep='') 
-				else x <- paste('dict([', paste(x, collapse=', '), '])', sep='')
+				if (any(duplicated(nms))) x <- paste('[', paste(x, collapse=','), ']', sep='') 
+				else x <- paste('dict([', paste(x, collapse=','), '])', sep='')
 				}
 			}
 		else
-			x <- paste('[', paste(x, collapse=', '), ']', sep='')
+			x <- paste('[', paste(x, collapse=','), ']', sep='')
 		return(x) }
 	Str4Py <- function(x, outmost=FALSE) {
 		# no considering on NA, Inf, ...
@@ -381,12 +503,14 @@ class R: # (object):
 	''' 
 	_DEBUG_MODE = True
 
-	def __init__(self, RCMD='R', max_len=1000, use_numpy=True, use_dict=None, host='localhost', user=None, ssh='ssh', return_err=True):
+	def __init__(self, RCMD='R', max_len=1000, use_numpy=True, use_pandas=True, use_dict=None, host='localhost', user=None, ssh='ssh', return_err=True):
 		'''
 		RCMD: The name of a R interpreter, path information should be included
 			if it is not in the system search path.
 		use_numpy: Used as a boolean value. A False value will disable numpy
 			even if it has been imported.
+		use_pandas: Used as a boolean value. A False value will disable pandas
+			even if it has been imported.
 		use_dict: A R named list will be returned as a Python dictionary if
 			"use_dict" is True, or a list of tuples (name, value) if "use_dict"
 			is False. If "use_dict" is None, the return value will be a
@@ -433,15 +557,32 @@ class R: # (object):
 			if arg not in RCMD: RCMD.append(arg)
 		if _has_subp and hasattr(subprocess, 'STARTUPINFO'):
 			info = subprocess.STARTUPINFO()
-			info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
-			info.wShowWindow = subprocess.SW_HIDE
+			try:
+				if hasattr(subprocess, '_subprocess'):
+					info.dwFlags |= subprocess._subprocess.STARTF_USESHOWWINDOW
+					info.wShowWindow = subprocess._subprocess.SW_HIDE
+				else:
+					info.dwFlags |= subprocess.STARTF_USESHOWWINDOW
+					info.wShowWindow = subprocess.SW_HIDE
+			except: info = None
 		else: info = None
+		# create stderr to replace None for py2exe: 
+		# http://www.py2exe.org/index.cgi/Py2ExeSubprocessInteractions
+		if sys.platform != 'win32':
+			childstderr = None
+		else:
+			if hasattr(sys.stderr, 'fileno'): 
+				childstderr = sys.stderr
+			elif hasattr(sys.stderr, '_file') and hasattr(sys.stderr._file, 'fileno'):
+				childstderr = sys.stderr._file
+			else: # Give up and point child stderr at nul 
+				childstderr = file('nul', 'a')
+                
 		self.__dict__.update({
-			'prog' : Popen(RCMD, stdin=PIPE, stdout=PIPE, stderr=return_err and _STDOUT or None, startupinfo=info), 
-			'has_numpy' : use_numpy and has_numpy, 
-			'Rfun' : self.__class__.__Rfun % ((use_numpy and has_numpy) and 'TRUE' or 'FALSE')})
+			'prog' : Popen(RCMD, stdin=PIPE, stdout=PIPE, stderr=return_err and _STDOUT or childstderr, startupinfo=info), 
+			'has_numpy' : use_numpy and has_numpy, 'has_pandas' : use_pandas and has_pandas, 
+			'Rfun' : self.__class__.__Rfun })
 		self.__call__(self.Rfun)
-		#to_discard = recv_some(self.prog, e=0, t=wait0)
 
 	def __runOnce(self, CMD, use_try=None):
 		'''
@@ -452,7 +593,8 @@ class R: # (object):
 		tail_token = 'R command at time: %s' % repr(time.time())
 		#tail_token_r = re.sub(r'[\(\)\.]', r'\\\1', tail_token)
 		tail_cmd = 'print("%s")%s' % (tail_token, newline)
-		re_tail = re.compile(r'>\sprint\("%s"\)\r?\n\[1\]\s"%s"\r?\n$' % (tail_token.replace(' ', '\\s'), tail_token.replace(' ', '\\s')) )
+		tail_token = tail_token.replace(' ', '\\s').replace('.', '\\.').replace('+', '\\+')
+		re_tail = re.compile(r'>\sprint\("%s"\)\r?\n\[1\]\s"%s"\r?\n$' % (tail_token, tail_token) )
 		if len(CMD) <= self.max_len or not self.localhost: 
 			fn = None
 		else:
@@ -460,7 +602,8 @@ class R: # (object):
 			os.fdopen(fh, 'wb').write(_mybytes(CMD))
 			if sys.platform == 'cli': os.close(fh) # this is necessary on IronPython 
 			CMD = 'source("%s")' % fn.replace('\\', '/')
-		CMD = (use_try and 'try({%s})%s%s' or '%s%s%s') % (CMD, newline, tail_cmd)
+		#CMD = (use_try and 'try({%s})%s%s' or '%s%s%s') % (CMD, newline, tail_cmd)
+		CMD = (use_try and 'try({%s})%s%s' or '%s%s%s') % (CMD.replace('\\', '\\\\'), newline, tail_cmd)
 		sendAll(self.prog, CMD)
 		rlt = ''
 		while not re_tail.search(rlt): 
@@ -485,12 +628,13 @@ class R: # (object):
 		if isinstance(CMDS, basestring): # a single command
 			rlt.append(self.__runOnce(CMDS, use_try=use_try))
 		else: # should be a list of commands
-			for CMD in CMDS: 
-				rlt.append(self.__runOnce(CMD, use_try=use_try))
+			#for CMD in CMDS: 
+			#	rlt.append(self.__runOnce(CMD, use_try=use_try))
+			rlt.append(self.__runOnce('; '.join(CMDS), use_try=use_try)) # now, ['sink("output.txt")', ..., 'sink()'] is allowed!
 		if len(rlt) == 1: rlt = rlt[0]
 		return rlt
 
-	def __getitem__(self, obj, use_try=None, use_dict=None): # to model r['XXX']
+	def __getitem__(self, obj, use_try=None, use_dict=None): # to model a dict: "r['XXX']"
 		'''
 		Get the value of an R variable or expression. The return value is
 		converted to the corresponding Python object.
@@ -505,7 +649,7 @@ class R: # (object):
 			raise RError('Leading underscore ("_") is not permitted in R variable names!')
 		use_try = use_try or self._DEBUG_MODE
 		if use_dict is None: use_dict = self.use_dict
-		cmd = '.getRvalue4Python__(%s, use_dict=%s)' % (obj, use_dict is None and 'NULL' or use_dict and 'TRUE' or 'FALSE')
+		cmd = '.getRvalue4Python__(%s, use_dict=%s, has_numpy=%s, has_pandas=%s)' % (obj, use_dict is None and 'NULL' or use_dict and 'TRUE' or 'FALSE', self.has_numpy and 'TRUE' or 'FALSE', self.has_pandas and 'TRUE' or 'FALSE')
 		rlt = self.__call__(cmd, use_try=use_try)
 		head = (use_try and 'try({%s})%s[1] ' or '%s%s[1] ') % (cmd, self.newline) 
 		# sometimes (e.g. after "library(fastICA)") the R on Windows uses '\n' instead of '\r\n'
@@ -517,7 +661,7 @@ class R: # (object):
 			raise RError(rlt)
 		return rlt
 
-	def __setitem__(self, obj, val): # to model r['XXX']
+	def __setitem__(self, obj, val): # to model a dict: "r['XXX'] = YYY"
 		'''
 		Assign a value (val) to an R variable (obj).
 
@@ -528,16 +672,16 @@ class R: # (object):
 			raise RError('Leading underscore ("_") is not permitted in R variable names!')
 		self.__call__('%s <- %s' % (obj, Str4R(val)))
 
-	def __delitem__(self, obj):
+	def __delitem__(self, obj): # to model a dict: "del r['XXX']"
 		if obj.startswith('_'):
 			raise RError('Leading underscore ("_") is not permitted in R variable names!')
 		self.__call__('rm(%s)' % obj)
 
-	def __del__(self):
+	def __del__(self): # to model "del r"
 		sendAll(self.prog, 'q("no")'+self.newline)
 		self.prog = None
 
-	def __getattr__(self, obj, use_dict=None): # to model r.XXX
+	def __getattr__(self, obj, use_dict=None): # to model object attribute: "r.XXX"
 		'''
 		obj: a string - the name of an R variable
 		use_dict: named list will be returned a dict if use_dict is True,
@@ -553,20 +697,20 @@ class R: # (object):
 			raise RError('No this object!')
 		return rlt
 
-	def __setattr__(self, obj, val): # to model r.XXX
+	def __setattr__(self, obj, val): # to model object attribute: "r.XXX = YYY"
 		if obj in self.__dict__ or obj in self.__class__.__dict__: # or obj.startswith('_'):
 			self.__dict__[obj] = val # for old-style class
 			#object.__setattr__(self, obj, val) # for new-style class
 		else:
 			self.__setitem__(obj, val)
 
-	def __delattr__(self, obj):
+	def __delattr__(self, obj): # to model object attribute: "del r.XXX"
 		if obj in self.__dict__:
 			del self.__dict__[obj]
 		else:
 			self.__delitem__(obj)
 
-	def get(self, obj, default=None, use_dict=None):
+	def get(self, obj, default=None, use_dict=None): # to model a dict: "r.get('XXX', 'YYY')"
 		'''
 		obj: a string - the name of an R variable, or an R expression
 		default: a python object - the value to be returned if failed to get data from R
@@ -587,7 +731,7 @@ class R: # (object):
 
 
 # for a single-round duty:
-def runR(CMDS, Robj='R', max_len=1000, use_numpy=True, use_dict=None, host='localhost', user=None, ssh='ssh'):
+def runR(CMDS, Robj='R', max_len=1000, use_numpy=True, use_pandas=True, use_dict=None, host='localhost', user=None, ssh='ssh'):
 	'''
 	Run a (list of) R command(s), and return the output from the STDOUT.
 
@@ -598,6 +742,8 @@ def runR(CMDS, Robj='R', max_len=1000, use_numpy=True, use_dict=None, host='loca
 		than this value.
 	use_numpy: Used as a boolean value. A False value will disable numpy even
 		if it has been imported.
+	use_pandas: Used as a boolean value. A False value will disable pandas even
+		if it has been imported.
 	use_dict: named list will be returned a dict if use_dict is True, otherwise
 		it will be a list of tuples (name, value).
 	host: The computer name (or IP) on which the R interpreter is
@@ -614,7 +760,7 @@ def runR(CMDS, Robj='R', max_len=1000, use_numpy=True, use_dict=None, host='loca
 	ssh: The program to login to remote computer.
 	'''
 	if isinstance(Robj, basestring):
-		Robj = R(RCMD=Robj, max_len=max_len, use_numpy=use_numpy, use_dict=use_dict, host=host, user=user, ssh=ssh)
+		Robj = R(RCMD=Robj, max_len=max_len, use_numpy=use_numpy, use_pandas=use_pandas, use_dict=use_dict, host=host, user=user, ssh=ssh)
 	rlt = Robj.run(CMDS=CMDS)
 	if len(rlt) == 1: rlt = rlt[0]
 	return rlt
